//---------------------------------------------------------
// Clock Speed Hack
// 2021.11.11 RSP
//
// Target: Pro Mini 3.3V/8MHz
//       + DS3231 RTC or GPS module PPS
//       + Common DIY clock movement
//
// Runs the clock movement at custom speed.
// Precision timebase can be any precision 1 second signal
//   such as DS3231 RTC SQW or GPS 1Hz PPS.
// Clock movement is slaved to the precision timebase.
//---------------------------------------------------------

// Pro Mini pins
#define SQW_PIN   3 // 1Hz interrupt
#define COIL_PIN1 8 // clock movement
#define COIL_PIN2 9 // clock movement

// clock configuration

#if (1)
// Mars clock
// expected 1027.4912510416665 msec for perfect martian second
// fractional part is stored to full precision using 64 bit math
#define SYNTH_SEC_MS   1027 // approximation of earth msec in mars second
#define SYNTH_SEC_FRAC 4912510416665LL // fractional part
#define SYNTH_HALF     5000000000000LL // 1/2
#define SYNTH_ONE     10000000000000LL // integer part
#define MOTOR_SPEED_BW 50   // motor speed limit (msec)
#endif

#if (0)
// Earth (standard) clock
#define SYNTH_SEC_MS   1000 // approximation of earth msec in clock second
#define SYNTH_SEC_FRAC 0LL
#define SYNTH_HALF     5LL
#define SYNTH_ONE     10LL
#define MOTOR_SPEED_BW 50
#endif

#if (0)
// 24 Hour Earth clock (1/2 speed)
#define SYNTH_SEC_MS   2000 // approximation of earth msec in clock second
#define SYNTH_SEC_FRAC 0LL
#define SYNTH_HALF     5LL
#define SYNTH_ONE     10LL
#define MOTOR_SPEED_BW 100
#endif

//---------------------------------------------------------
// Precision timebase

// https://github.com/hideakitai/DS323x
#include <DS323x.h>
DS323x rtc;

// interrupt counts precision msecs
volatile int      precision_ms_count = 0; // msec counter
volatile uint32_t precision_ms_tm    = 0; // msec tm
volatile uint16_t dbg_precision_sec_count = 0; // diagnostic

void int1Hz() // 1 Hz precision timebase (SQW signal)
{
  // count precision msecs  
  precision_ms_count += 1000;
  precision_ms_tm = millis();

  // diagnostic; heartbeat
  digitalWrite( LED_BUILTIN, !digitalRead(LED_BUILTIN) ); 

  // diagnostic; count elapsed precision seconds
  dbg_precision_sec_count++;
}

void sync() // sync with SQW edge
{
Serial.print("waiting timebase sync...");  
  digitalWrite( LED_BUILTIN, HIGH ); 
  while (!digitalRead(SQW_PIN)); // wait for HIGH
  while ( digitalRead(SQW_PIN)); // wait for LOW
  dbg_precision_sec_count = 0; 
  precision_ms_count = 0;
  precision_ms_tm = millis(); 
Serial.println("started!");
}

//---------------------------------------------------------
// Program start

void setup() 
{
  // default unused pins
  for (uint8_t i=0; i <= 17; i++) pinMode(i,INPUT_PULLUP);  

  // set up alternating phase clock motor drive outputs
  digitalWrite(COIL_PIN1,LOW); pinMode( COIL_PIN1,OUTPUT );
  digitalWrite(COIL_PIN2,LOW); pinMode( COIL_PIN2,OUTPUT );

  // diagnostics
  Serial.begin(115200);
  pinMode(LED_BUILTIN,OUTPUT); // SQW signal heartbeat

  // enable precision 1Hz on RTC SQW pin
  Wire.begin();
  rtc.attach(Wire);
  // SQW is calibrated + temperature compensated frequency
  rtc.interruptControl(DS323x::InterruptCtrl::SQW);
  // DS3231M is always 1Hz SQW; other variants require this setting
  rtc.squareWaveFrequency(DS323x::SquareWaveFreq::SQWF_1_HZ);
  // 32KHz is uncompensated, turn it off
  rtc.enable32kHz(false); // conserve power
  
  // set up 1Hz interrupt    
  pinMode(SQW_PIN, INPUT_PULLUP);
  attachInterrupt(digitalPinToInterrupt(SQW_PIN), int1Hz, FALLING);
  
  // sync with SQW interrupt edge to kick off control loop
  sync();
}

//---------------------------------------------------------
// Bigloop

void loop() // runs at ~1 clock sec/loop (continuously error corrected)
{
static boolean  dbg = false;              // enable verbose output to serial monitor
static uint32_t dbg_tm = millis();        // elapsed millis()

  // Count clock seconds (tocks, in msecs) and precision elapsed actual msecs (ticks, from 1Hz timebase)
  //   speed up or slow down clock motor based on error term so they match
  //   tick_error is the difference between expected and actual time
         uint16_t tot_ms = SYNTH_SEC_MS; // motor run time for this clock second
  static int16_t  ms_error = 0;          // 2'nd order error term
  static int64_t  tock_frac  = 0;        // expected msec fractional accumulator
  static int16_t  tick_error = 0;        // actual vs expected error term (msec)
  static uint16_t est_ms = 0;            // estimated part of tick_error

  // apply error terms
  tot_ms -= tick_error; // control loop objective is to keep tick_error at zero
  //  2'nd order error term compensates for loop overhead
  tot_ms   -= ms_error / 64;
  ms_error += tick_error; // x64 error accumulator 

  // sanity limit on clock motor speed range
  if (tot_ms < (SYNTH_SEC_MS-MOTOR_SPEED_BW/2)) tot_ms = SYNTH_SEC_MS-MOTOR_SPEED_BW/2;
  if (tot_ms > (SYNTH_SEC_MS+MOTOR_SPEED_BW/2)) tot_ms = SYNTH_SEC_MS+MOTOR_SPEED_BW/2;  

  // indicate error if out of control band
  if (tick_error < -MOTOR_SPEED_BW || tick_error > MOTOR_SPEED_BW)
  { // transient recovery
Serial.println("OUT OF CONTROL"); 
    sync();  ms_error = tock_frac = tick_error = est_ms = 0; dbg_tm = millis(); tot_ms = SYNTH_SEC_MS;
  }

  // break tot_ms into motor control pulses spread out evenly
#define OFF_MSEC 16 // motor coil off time
  uint16_t on_ms = tot_ms - OFF_MSEC*2*8; // total coil on time
  uint16_t pulse_ms  =  on_ms >> 4;       // 8 revolutions/sec, 2 cycles/revolution
  uint8_t  remainder = (on_ms & 15);
  // clock movement driver
  for (uint8_t i=0; i < 8; i++)
  { // 8 revolutions per second
    digitalWrite(COIL_PIN1,HIGH);
    delay(pulse_ms+(remainder ? 1:0)); if (remainder) remainder--;
    digitalWrite(COIL_PIN1,LOW);
    delay(OFF_MSEC);
    digitalWrite(COIL_PIN2,HIGH);
    delay(pulse_ms+(remainder ? 1:0)); if (remainder) remainder--;
    digitalWrite(COIL_PIN2,LOW);
    delay(OFF_MSEC);
  }

  // ticks: hybrid precision + millis() timebase
  //   count actual msecs based on precision 1Hz interrupt
  //   plus estimation using millis()
  //   ultimately tracks precision 1Hz timebase to keep clock accurate
  uint16_t prev_est_ms = est_ms;
  noInterrupts();
    uint32_t tm_now = millis();
    // estimated ms is time from last precision second to now.
    //   estimated ms will be removed in next cycle,
    //   leaving only precision msec being accumulated into tick_error.
    est_ms = tm_now - precision_ms_tm; // millis() estimate
    // ticks is precision (1Hz int) msec + estimated (millis) msec
    uint16_t ticks = precision_ms_count + est_ms - prev_est_ms;
    precision_ms_count = 0;
    precision_ms_tm = tm_now;
  interrupts();

  // tocks: expected time for one clock second
  //   tock accumulator is integer part (tocks) plus decimal part (tock_frac)
  uint16_t tocks = SYNTH_SEC_MS;      // integer part
         tock_frac += SYNTH_SEC_FRAC; // accumulate error term
  while (tock_frac >  SYNTH_HALF)     // when term exceeds 1/2
      {  tock_frac -= SYNTH_ONE;      //   carry 1 to integer part
         tocks++;
      }
  
  // reconcile expected vs actual msec to error term
  tick_error += ticks - tocks;

// diagnostic monitor; enter any printable character to toggle verbose debug output
int c=Serial.read(); if (c > 0x20 && c < 0x7F) dbg = !dbg;
if (dbg)
{ Serial.print(dbg_precision_sec_count); Serial.write(' '); // actual time (precision timer)
  Serial.print(millis()-dbg_tm); Serial.write(' ');         // millis() time
  Serial.print(tot_ms); Serial.write(' ');                  // this period (msec)
  Serial.print(ms_error); Serial.write(' ');                // 2'nd order error term (loop stability)
  Serial.print(ticks); Serial.write(' ');                   // actual time (msec) this period
  Serial.print(tocks); Serial.write(' ');                   // ideal time (msec) this period
  Serial.println(tick_error);                               // cumulative error (msec)
}
    
}
